"""
Этот модуль содержит сортировки.
"""

from .heap import MinHeap
from .my_functions import swap, merge


def heap_sort(arr: list, **kwargs):
    """
    Сортирует входной список с использованием алгоритма сортировки кучей.

    .. image:: images/heapsort.gif

    :Сложность: O(N log N)
    :param arr: Список, который требуется отсортировать.
    :param kwargs: Дополнительные параметры для настройки поведения сортировки.
                   Возможные параметры:

       - reverse (bool): Если установлено в True, возвращает список в обратном порядке.
       - inplace (bool): Если установлено в True, сортирует список на месте (in-place).

    :return: Отсортированный список.
    """
    inplace = kwargs.get("inplace", False)
    reverse = kwargs.get("reverse", False)

    length = len(arr)
    if length <= 1:
        return arr

    if not inplace:
        arr = arr.copy()

    heap = MinHeap(arr)
    for i in range(length):
        arr[i] = heap.poll()

    if reverse:
        arr.reverse()

    return arr


def bubble_sort(arr: list, **kwargs):
    """
    Сортировка пузырьком.

    .. image:: images/bubble-sort.gif

    :Сложность: O(N²)
    :param arr: Список, который требуется отсортировать.
    :param kwargs: Дополнительные параметры для настройки поведения сортировки.
                   Возможные параметры:

                   - reverse (bool): Если установлено в True, возвращает список в обратном порядке.
                   - inplace (bool): Если установлено в True, сортирует список на месте (in-place).

    :return: Отсортированный список.
    """
    inplace = kwargs.get("inplace", False)
    reverse = kwargs.get("reverse", False)

    if len(arr) <= 1:
        return arr

    if not inplace:
        arr = arr.copy()

    for i in range(len(arr) - 1):
        for j in range(i + 1, len(arr)):
            if arr[i] > arr[j]:
                swap(arr, i, j)

    if reverse:
        arr.reverse()

    return arr


def quicksort(arr, left=None, right=None, **kwargs):
    """
    Выполняет алгоритм быстрой сортировки на входном списке.

    .. image:: images/quicksort.gif

    :Сложность: O(N log N)
    :param arr: Список для сортировки.
    :param left: Начальный индекс подсписка для сортировки. По умолчанию 0.
    :param right: Конечный индекс подсписка для сортировки. По умолчанию len(arr) - 1.

    :param kwargs: Дополнительные параметры для поведения сортировки.

       - inplace (bool): Если True, сортирует список на месте. По умолчанию False.
       - reverse (bool): Если True, сортирует список в обратном порядке. По умолчанию False.

    :return: Отсортированный список.
    """
    if len(arr) <= 1:
        return arr

    if left is None:
        left = 0
    if right is None:
        right = len(arr) - 1

    inplace = kwargs.get("inplace", False)
    reverse = kwargs.get("reverse", False)

    if not inplace:
        arr = arr.copy()

    if left < right:
        sep_index = _partition(arr, left, right, reverse)
        quicksort(arr, left, sep_index - 1, reverse=reverse, inplace=True)
        quicksort(arr, sep_index + 1, right, reverse=reverse, inplace=True)

    return arr


def _partition(arr, left, right, reverse):
    """
    Эта функция реализует логику разделения для алгоритма быстрой сортировки.
    Она выбирает последний элемент в качестве опорного и переставляет массив так,
    чтобы все элементы меньше опорного были слева, а все элементы больше опорного - справа,
    либо наоборот, если reverse == True.

    :param arr: Список для сортировки.
    :param left: Начальный индекс подсписка для сортировки.
    :param right: Конечный индекс подсписка для сортировки.
    :param reverse: Если True, сортирует список в обратном порядке. \
    Если False, сортирует в прямом порядке.

    :return: Индекс опорного элемента после разделения.
    """
    pivot = arr[right]
    i = left - 1

    for j in range(left, right):
        if not reverse and arr[j] < pivot or reverse and arr[j] > pivot:
            i += 1
            swap(arr, i, j)

    swap(arr, i + 1, right)
    return i + 1

def merge_sort(arr: list, **kwargs) -> list:
    """
    Сортировка слиянием.

    .. image:: images/merge.gif

    :Сложность: O(N log N)
    :param arr: Список, который требуется отсортировать.
    :param kwargs: Дополнительные параметры для настройки поведения сортировки.
                   Возможные параметры:

                   - reverse (bool): Если установлено в True, возвращает список в обратном порядке.
                   - inplace (bool): Если установлено в True, сортирует список на месте (in-place).

    :return: Отсортированный список.
    """
    if len(arr) <= 1:
        return arr

    inplace = kwargs.get("inplace", False)
    reverse = kwargs.get("reverse", False)

    mid = len(arr) // 2
    left_half = merge_sort(arr[:mid])
    right_half = merge_sort(arr[mid:])
    sorted_arr = merge(left_half, right_half)

    if reverse:
        sorted_arr.reverse()

    if inplace:
        arr[:] = sorted_arr
        return arr

    return sorted_arr


def insertion_sort(arr: list, **kwargs) -> list:
    """
    Сортировка вставкой.

    .. image:: images/insertion.gif

    :Сложность: O(N²)
    :param arr: Список, который требуется отсортировать.
    :param kwargs: Дополнительные параметры для настройки поведения сортировки.
                   Возможные параметры:

                   - reverse (bool): Если установлено в True, возвращает список в обратном порядке.
                   - inplace (bool): Если установлено в True, сортирует список на месте (in-place).

    :return: Отсортированный список.
    """
    if len(arr) <= 1:
        return arr

    inplace = kwargs.get("inplace", False)
    reverse = kwargs.get("reverse", False)

    if not inplace:
        arr = arr.copy()

    for i in range(1, len(arr)):
        j = i - 1
        while j >= 0 and (arr[j] > arr[j + 1] if not reverse else arr[j] < arr[j + 1]):
            swap(arr, j, j + 1)
            j -= 1

    return arr


def selection_sort(arr: list, **kwargs) -> list:
    """
    Сортировка выборкой.

    .. image:: images/selection.gif

    :Сложность: O(N²)
    :param arr: Список, который требуется отсортировать.
    :param kwargs: Дополнительные параметры для настройки поведения сортировки.
                   Возможные параметры:

                   - reverse (bool): Если установлено в True, возвращает список в обратном порядке.
                   - inplace (bool): Если установлено в True, сортирует список на месте (in-place).

    :return: Отсортированный список.
    """
    if len(arr) <= 1:
        return arr

    inplace = kwargs.get("inplace", False)
    reverse = kwargs.get("reverse", False)

    if not inplace:
        arr = arr.copy()

    for i in range(len(arr)):
        mini = i
        for j in range(i+1, len(arr)):
            if arr[j] < arr[mini]:
                mini = j
        swap(arr, i, mini)

    if reverse:
        arr.reverse()

    return arr



"""
Этот модуль содержит алгоритмы, связанные с работой со строками.
"""

def brute_force_search(text: str, substring: str) -> int:
    """
    Грубый поиск подстроки.

    .. image:: images/brute-force-string-search.gif
        :width: 400px

    :Сложность: O(N * M), где N - длина строки, а M - длина подстроки
    :param text: Исходный текст.
    :param substring: искомая подстрока. 
    
    :return: Индекс начала подстроки в строке (если не найдено, то -1).
    """

    n = len(text)
    m = len(substring)

    if m == 0:
        return -1

    for i in range(n - m + 1):
        j = 0
        while j < m and text[i + j] == substring[j]:
            j += 1
        if j == m:
            return i
    return -1


def _calculate_hash(s, length, base, prime):
    """
    Функция, создающая хэш подстроки. Она создает хэш плавающего окна.
    """
    h = 0
    for i in range(length):
        h = (base * h + ord(s[i])) % prime
    return h

def _update_hash(current_hash, old_char, new_char, h, base, prime):
    """
    Функция, обновляющая хэш подстроки. Она сдвигает плавающее окно вправо.
    """
    current_hash = (base * (current_hash - ord(old_char) * h) + ord(new_char)) % prime
    if current_hash < 0:
        current_hash += prime
    return current_hash

def rabin_karp(text: str, substring: str, base=256, prime=89) -> int:
    """
    Алгоритм Рабина-Карпа создан для поиска подстроки в строке за линейное время.
    Вместо того, чтобы сравнивать строки, мы сравниваем из хэши, что позволяет
    проверять совпадают ли строки за O(1).

    .. image:: images/rabin-karp.gif
        :width: 400px

    :Сложность: O(N + M), где N - длина строки, а M - длина подстроки
    :param text: Исходный текст.
    :param substring: искомая подстрока.
    :param base: количество символов в алфавите строки. По умолчанию 256 (ASCII).
    :param prime: простое число для генерации хэша. По умолчанию 89.

    :return: Индекс начала подстроки в строке (если не найдено, то -1).
    """

    n = len(text)
    m = len(substring)

    if n < m or m == 0:
        return -1

    h = 1 # значение хэша для сдвига вправо
    for i in range(m - 1):
        h = (h * base) % prime

    substring_hash = _calculate_hash(substring, m, base, prime)
    current_hash = _calculate_hash(text, m, base, prime)

    for i in range(n - m + 1):
        if substring_hash == current_hash:
            if text[i:i + m] == substring:
                return i

        if i < n - m:
            current_hash = _update_hash(current_hash, text[i], text[i + m], h, base, prime) # O(1)

    return -1



"""
Этот модуль содержит различные узлы бинарных деревьев.
Эти узы используются в , но их также можно использовать отдельно.
У узлов :class:`~BinaryNode`
и :class:`~BalancingNode` нет своих деревьев, но их можно использовать отдельно.
**Остальные узлы использовать отдельно не рекомендуется**.
"""


from abc import ABC, abstractmethod
from ..linear import Queue


def height(node) -> int:
    """
    Возвращает высоту входящего узла. Сложность алгоритма является **O(1)**
    вместо O(h), потому что каждый узел хранит в себе значение его высоты.
    При каждом изменении дерева это значение обновляется при необходимости.

    :Сложность: O(1)
    :param node: узел, высоту которого надо узнать
    :return: высота узла и, если узла нет, -1
    """
    if node:
        return node.height
    return -1


class Node(ABC):
    """
    Абстрактный класс, который описывает основные свойства узлов дерева.
    """

    @property
    @abstractmethod
    def parent(self):
        """
        Родитель элемента.
        """
        raise NotImplementedError

    @parent.setter
    @abstractmethod
    def parent(self, value):
        raise NotImplementedError

    @property
    @abstractmethod
    def left(self):
        """
        Левый ребёнок элемента.
        """
        raise NotImplementedError

    @left.setter
    @abstractmethod
    def left(self, value):
        raise NotImplementedError

    @property
    @abstractmethod
    def right(self):
        """
        Правый ребёнок элемента.
        """
        raise NotImplementedError

    @right.setter
    @abstractmethod
    def right(self, value):
        raise NotImplementedError

    @abstractmethod
    def maintain(self):
        """
        Обработка дерева после определённых действий (при необходимости).
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def data(self):
        """
        Данные, хранящиеся внутри узла.
        """
        raise NotImplementedError

    @data.setter
    @abstractmethod
    def data(self, value):
        raise NotImplementedError


class BinaryNode(Node):
    """
    Узел бинарного дерева.
    """

    def __init__(self, item):
        """
        Инициализатор.

        :param item: Значение узла
        """
        self._data = item
        self._parent = None
        self._right = None
        self._left = None

    def __repr__(self):
        return f"Node({self.data})"

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def left(self):
        return self._left

    @left.setter
    def left(self, value):
        self._left = value

    @property
    def right(self):
        return self._right

    @right.setter
    def right(self, value):
        self._right = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    def maintain(self):
        """
        Функция вызывается после удаления или вставки узла (*уже прописано в коде*).
        Заготовка.
        """

    def subtree_iter(self):
        """
        Рекурсивный обход дерева `по порядку` (in-order traversal), также `центрированный`.
        Является одним из трёх обходов в глубину (**DFS** - `Depth First Search`).

        Алгоритм обхода по порядку:
            #. Обойти левое *поддерево* (рекурсивно).
            #. *Выбросить* (оператор yield) *корень*.
            #. Обойти правое *поддерево* (рекурсивно).

        .. image:: images/inorder-traversal.gif
            :width: 400px

        :Сложность: O(n)
        :return: Итерационный объект, состоящий из узлов бинарного дерева по порядку.
        """
        if self.left:
            yield from self.left.subtree_iter()
        yield self
        if self.right:
            yield from self.right.subtree_iter()

    def inorder_traversal(self):
        """
        Рекурсивный обход дерева `по порядку` (in-order traversal), также `центрированный`.
        Является одним из трёх обходов в глубину (**DFS** - `Depth First Search`).

        Смотри также:

        * :class:`~BinaryTreeNode.preorder_traversal`
        * :class:`~BinaryTreeNode.postorder_traversal`

        Обход дерева по умолчанию (:class:`~BinaryTreeNode.subtree_iter`.).

        Алгоритм обхода по порядку:
            #. Обойти левое *поддерево* (рекурсивно).
            #. *Выбросить* (оператор yield) *корень*.
            #. Обойти правое *поддерево* (рекурсивно).

        .. image:: images/inorder-traversal.gif
            :width: 400px

        :Сложность: O(n)
        :return: Итерационный объект, состоящий из узлов бинарного дерева по порядку.
        """
        return self.subtree_iter()

    def preorder_traversal(self):
        """
        Рекурсивный обход дерева `по порядку` (preorder traversal), также `прямой`.
        Является одним из трёх обходов в глубину (**DFS** - `Depth First Search`).

        Смотри также:

        * :class:`~BinaryTreeNode.inorder_traversal`
        * :class:`~BinaryTreeNode.postorder_traversal`

        Алгоритм прямого обхода:
            #. *Выбросить* (оператор yield) *корень*.
            #. Обойти левое *поддерево* (рекурсивно).
            #. Обойти правое *поддерево* (рекурсивно).

        .. image:: images/preorder-traversal.gif
            :width: 400px

        :Сложность: O(n)
        :return: Итерационный объект, состоящий из узлов бинарного дерева.
        """
        yield self
        if self.left:
            yield from self.left.preorder_traversal()
        if self.right:
            yield from self.right.preorder_traversal()

    def postorder_traversal(self):
        """
        Рекурсивный обход дерева `по порядку` (postorder traversal), также `обратный`.
        Является одним из трёх обходов в глубину (**DFS** - `Depth First Search`).

        Смотри также:

        * :class:`~BinaryTreeNode.preorder_traversal`
        * :class:`~BinaryTreeNode.inorder_traversal`

        Алгоритм обратного обхода:
            #. Обойти левое *поддерево* (рекурсивно).
            #. Обойти правое *поддерево* (рекурсивно).
            #. *Выбросить* (оператор yield) *корень*.

        .. image:: images/postorder-traversal.gif
            :width: 400px

        :Сложность: O(n)
        :return: Итерационный объект, состоящий из узлов бинарного дерева.
        """
        if self.left:
            yield from self.left.postorder_traversal()
        if self.right:
            yield from self.right.postorder_traversal()
        yield self

    def level_order_traversal(self):
        """
        Обход в ширину (**BFS** - `Breadth First Search`).

        Алгоритм обхода в ширину:
            #. *Выбросить* (оператор yield) *корень*.
            #. Обойти следующий слой

        .. image:: images/level-order-traversal.gif
            :width: 400px

        :Сложность: O(n)
        :return: Итерационный объект, состоящий из узлов бинарного дерева.
        """

        queue = Queue()
        queue.enqueue(self)
        visited = set()

        while queue:
            node = queue.dequeue()
            visited.add(node)

            yield node

            if node.left and node.left not in visited:
                queue.enqueue(node.left)
            if node.right and node.right not in visited:
                queue.enqueue(node.right)

    def subtree_first(self):
        """
        Получить первый по порядку узел дерева (самый левый).

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: Первый узел дерева
        """
        if self.left:
            return self.left.subtree_first()
        return self

    def subtree_last(self):
        """
        Получить последний по порядку узел дерева (самый правый).

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: Последний узел дерева
        """
        if self.right:
            return self.right.subtree_last()
        return self

    def successor(self):
        """
        Найти преемника этого узла (узел, который идёт следующий по порядку).

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: Преемник этого узла
        """
        if self.right:
            return self.right.subtree_first()

        while self.parent and (self is self.parent.right):
            self = self.parent

        return self

    def predecessor(self):
        """
        Найти предшественника этого узла (узел, который идёт предыдущий по порядку).

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: Предшественник этого узла
        """
        if self.left:
            return self.left.subtree_last()
        while self.parent and (self is self.parent.left):
            self = self.parent

        return self

    def subtree_insert_before(self, node: Node):
        """
        Вставить узел (node) перед текущим по порядку.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param node: Узел, который мы хотим вставить перед текущим
        """
        if self.left:
            self = self.left.subtree_last()
            self.right, node.parent = node, self
        else:
            self.left = node
            node.parent = self

        self.maintain()

    def subtree_insert_after(self, node: Node):
        """
        Вставить узел (node) после текущего по порядку.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param node: Узел, который мы хотим вставить после текущего
        """
        if self.right:
            self = self.right.subtree_first()
            self.left, node.parent = node, self
        else:
            self.right, node.parent = node, self

        self.maintain()

    def subtree_delete(self) -> Node:
        """
        Рекурсивная функция удаления узла дерева.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: Удаляемый узел
        """
        if self.right or self.left:
            if self.left:
                node = self.predecessor()
            else:
                node = self.successor()
            node.data, self.data = self.data, node.data
            return node.subtree_delete()
        if self.parent:
            if self.parent.left is self:
                self.parent.left = None
            else:
                self.parent.right = None
            self.parent.maintain()
            self.parent = None

        return self


class BalancingNode(BinaryNode):
    """
    Узел балансирующего узла.
    """
    def __init__(self, item):
        super().__init__(item)
        self.height = None
        self.subtree_update()

    def subtree_rotate_left(self):
        """
        Функция поворота дерева налево нужна для поддержки баланса дерева.
        Она уменьшает скос дерева, не меняя при этом порядок узлов при обходе дерева по порядку.
        Функция вызывается относительно корня поддерева, которое надо повернуть.
        Функция вызывается при каждом обновлении дерева,
        см. :class:`~BinaryTreeNode.maintain`.

        **Алгоритм:**

        1. Обозначить временные переменные для хранения узлов: \
        *root_left_subtree*, *pivot_left_subtree*, *pivot_right_subtree*
        2. Поменять местами корень (*root*) и опорную точку (*pivot*). Теперь *pivot* - корень
        3. Сделать левым ребенком *pivot*'a корень (*root*), а правым -- *pivot_right_subtree*
        4. Сделать левым ребёнком *root*'a *root_left_subtree*, а правым -- *pivot_left_subtree*
        5. Не забыть поставить указатели на своих родителей для \
        *pivot_right_subtree* и *root_left_subtree*
        6. Обновить дерево относительно *pivot*'a и *root*'a

        .. image:: images/left_rotation.gif
            :width: 400px

        :Сложность: O(1)
        """
        assert self.right
        root_left_subtree, pivot = self.left, self.right
        pivot_left_subtree, pivot_right_subtree = pivot.left, pivot.right
        self, pivot = pivot, self
        self.data, pivot.data = pivot.data, self.data
        pivot.left, pivot.right = self, pivot_right_subtree
        self.left, self.right = root_left_subtree, pivot_left_subtree
        if root_left_subtree:
            root_left_subtree.parent = self
        if pivot_right_subtree:
            pivot_right_subtree.parent = pivot
        self.subtree_update()
        pivot.subtree_update()

    def subtree_rotate_right(self):
        """
        Функция поворота дерева направо нужна для поддержки баланса дерева.
        Она уменьшает скос дерева, не меняя при этом порядок узлов при
        обходе дерева по порядку. Функция вызывается относительно корня поддерева,
        которое надо повернуть. Функция вызывается при каждом обновлении дерева,
        см. :class:`~BinaryTreeNode.maintain`.

        **Алгоритм:**

        1. Обозначить временные переменные для хранения узлов: \
        *root_right_subtree*, *pivot_left_subtree*, *pivot_right_subtree*
        2. Поменять местами корень (*root*) и опорную точку (*pivot*). Теперь *pivot* - корень
        3. Сделать правым ребенком *pivot*'a корень (*root*), а левым -- *pivot_left_subtree*
        4. Сделать правым ребёнком *root*'a *root_right_subtree*, а левым -- *pivot_right_subtree*
        5. Не забыть поставить указатели на своих родителей для \
        *pivot_left_subtree* и *root_right_subtree*
        6. Обновить дерево относительно *pivot*'a и *root*'a

        .. image:: images/right_rotation.gif
            :width: 400px

        :Сложность: O(1)
        """
        assert self.left
        pivot, root_right_subtree = self.left, self.right
        pivot_left_subtree, pivot_right_subtree = pivot.left, pivot.right
        self, pivot = pivot, self
        self.data, pivot.data = pivot.data, self.data
        pivot.left, pivot.right = pivot_left_subtree, self
        self.left, self.right = pivot_right_subtree, root_right_subtree
        if pivot_left_subtree:
            pivot_left_subtree.parent = pivot
        if root_right_subtree:
            root_right_subtree.parent = self

        pivot.subtree_update()
        self.subtree_update()

    def subtree_update(self):
        """
        Обновляет высоту узла. *Высота самого высокого ребёнка + 1*.

        :Сложность: O(1)
        """
        self.height = 1 + max(height(self.left), height(self.right))

    def skew(self):
        """
        Найти *скос* дерева. *Скос* - это разница между высотами поддеревьев
        (*в данном случае высота правого минус высота левого*).
        Если эта разница больше нуля, то дерево накренено **вправо**.
        Если эта разница меньше нуля, то дерево накренено **влево**.
        Если эта разница равна нулю, то дерево **полное** (не накренено).

        :Сложность: O(1)
        :return: разница высот.
        """
        return height(self.right) - height(self.left)


    def maintain(self):
        """
        Функция вызывается после удаления или вставки узла (*уже прописано в коде*).
        Она поддерживает дерево сбалансированным за счёт :class:`~BalancingNode.rebalance`
        Также оно проходиться вверх по дереву с той же целью.

        :Сложность: O(log n)
        """
        self.rebalance()
        self.subtree_update()
        if self.parent:
            self.parent.maintain()

    def rebalance(self):
        """
        Сбалансировать дерево, если оно слишком накренено:

        * Если накренено вправо, повернуть налево
            * Перед этим сбалансировать правое поддерево (*повернуть направо*) при необходимости
        * Если накренено влево, повернуть направо
            * Перед этим сбалансировать левое поддерево (*повернуть налево*) при необходимости

        :Сложность: O(1)
        """
        if self.skew() == 2:
            if self.right.skew() < 0:
                self.right.subtree_rotate_right()
            self.subtree_rotate_left()

        elif self.skew() == -2:
            if self.left.skew() > 0:
                self.left.subtree_rotate_left()
            self.subtree_rotate_right()


class SearchNode(BinaryNode):
    """
    Узел бинарного дерева поиска.
    """
    def subtree_find(self, item: int):
        """
        Найти узел.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение искомого узла
        :return: искомый узел
        """
        if item < self.data:
            if self.left is not None:
                return self.left.subtree_find(item)

        elif item > self.data:
            if self.right is not None:
                return self.right.subtree_find(item)

        else:
            return self

        return None

    def subtree_find_next(self, item: int):
        """
        Найти следующий узел.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение, после которого идет значение искомого узла
        :return: узел, значение которого идёт после входного значения
        """
        if self.data <= item:
            # идти вправо, чтобы увеличить значение узла
            if self.right:
                return self.right.subtree_find_next(item)
            return None

        # когда значение узла больше:
        # самый левый элемент поддерева (если он есть),
        # чтобы максимально приблизиться к искомому значению (убывание)
        if self.left:
            node = self.left.subtree_find_next(item)
            if node:
                return node

        # если нет узлов в левом поддереве, вернуть текущий узел
        return self

    def subtree_find_prev(self, item: int):
        """
        Найти предыдущий узел.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение, перед которым идет значение искомого узла
        :return: узел, значение которого идёт перед входным значением
        """
        if self.data >= item:
            if self.left:
                return self.left.subtree_find_prev(item)
            return None

        if self.right:
            node = self.right.subtree_find_prev(item)
            if node:
                return node

        return self

    def subtree_insert(self, new_node: Node):
        """
        Добавить новый узел и не нарушить порядок возрастания при обходе по порядку.

        .. image:: images/bst_insert.gif

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param new_node: новый узел, который мы добавляем
        """
        if new_node.data < self.data:
            if self.left:
                self.left.subtree_insert(new_node)
            else:
                self.subtree_insert_before(new_node)

        elif new_node.data > self.data:
            if self.right:
                self.right.subtree_insert(new_node)
            else:
                self.subtree_insert_after(new_node)

        else:
            # if new_node.data == self.data:
            self.data = new_node.data
            return

        new_node.maintain()


class RedBlackNode(SearchNode):
    """
    Узел красно-чёрного дерева.
    """
    RED = True
    BLACK = False

    def __init__(self, item: int, color=RED):
        super().__init__(item)
        self.color = color

    @property
    def grandparent(self):
        """
        Найти прародителя.
        """
        if self.parent is None:
            return None
        return self.parent.parent

    @property
    def sibling(self):
        """
        Найти брата.
        """
        if self.parent is None:
            return None
        if self == self.parent.left:
            return self.parent.right
        return self.parent.left

    @property
    def uncle(self):
        """
        Найти дядю.
        """
        if self.parent is None:
            return None
        return self.parent.sibling


class SegmentNode(BalancingNode):
    """
    Индексируемый узел имеет параметр размера. Размер -- это сумма размеров детей + 1.
    Это нужно для индексации узлов.
    """
    def __init__(self, item):
        super().__init__(item)
        self.size = 1


    def subtree_update(self):
        """
        Обновляет размер узла (*это сумма размеров детей + 1*).

        :Сложность: O(1)
        """
        self.size = 1
        if self.left:
            self.size += self.left.size

        if self.right:
            self.size += self.right.size

        super().subtree_update()

    def subtree_at(self, i):
        """
        Найти i-ый узел. Поиск происходит через размеры узлов:

        * Если размер левого ребёнка больше индекса, найти i-ый узел в левом поддереве.
        * Если размер левого ребёнка меньше индекса, найти узел в правом поддереве с индексом: \
        *индекс - размер левого ребёнка - 1*.
        * Если размер левого ребенка равен индексу, вернуть текущий узел.

        :Сложность: O(log n)
        :param i: индекс искомого узла
        :return: искомый узел
        """
        assert i >= 0
        if self.left:
            left_size = self.left.size
        else:
            left_size = 0

        if i < left_size:
            return self.left.subtree_at(i)
        if i > left_size:
            return self.right.subtree_at(i - left_size - 1)

        return self


class AVLNode(BalancingNode, SearchNode):
    """
    Узел дерева AVL. AVL-дерево — это сбалансированное бинарное дерево поиска,
    в котором для каждого узла высота его левого и правого поддерева может отличаться
    не более чем на 1.

    Наследует функциональность как от BSTNode (для операций поиска и вставки),
    так и от BalancingNode (для поддержания баланса дерева).
    """


class TrieNode:
    """
    Узел префиксного дерева.
    """

    def __init__(self):
        """
        Инициализирует узел TrieNode пустым словарём дочерних узлов и флагом is_word,
        установленным в False.
        """
        self.children = {}
        self.is_word = False

    def __repr__(self):
        return str(self.children)

    def __contains__(self, char):
        """
        Проверяет, есть ли у узла TrieNode дочерний узел с указанным символом.

        :param char: Символ для проверки.
        :return: True, если у узла TrieNode есть дочерний узел с указанным символом, иначе False.
        """
        return char in self.children

    def __setitem__(self, key, value):
        """
        Устанавливает дочерний узел для узла TrieNode с указанным символом и значением.

        :param key: Символ для установки в качестве дочернего узла.
        :param value: Узел TrieNode для установки в качестве дочернего узла.
        """
        self.children[key] = value

    def __getitem__(self, key):
        """
        Получает дочерний узел узла TrieNode с указанным символом.

        :param key: Символ для получения дочернего узла.
        :return: Узел TrieNode-дочерний с указанным символом.
        """
        return self.children[key]

    def get_words(self, prefix=""):
        """
        Генерирует все слова в поддереве, корнем которого является этот узел TrieNode
        с указанным префиксом.

        :param prefix: Префикс для добавления к словам.
        :return: Генератор, выдающий слова с указанным префиксом.
        """
        if self.is_word:
            yield prefix
        for char, node in self.children.items():
            yield from node.get_words(prefix + char)


class TwoThreeTreeNode:
    """
    Узел 2-3 дерева.
    """
    def __init__(self, keys=None, children=None):
        """
        Инициализатор принимает списки значений и детей.
        """
        self.keys = keys if keys else []
        self.children = children if children else []

    def __repr__(self):
        return f"Node({self.keys})"

    def is_leaf(self):
        """
        Проверяет, является ли узел листом.
        """
        return len(self.children) == 0

    def is_full(self) -> bool:
        """
        Проверяет узел на полноту (3 элемента).
        """
        return len(self.keys) == 3


class HuffmanNode:
    """
    Узел дерева Хаффмана.
    """
    def __init__(self, char, freq):
        """
        Инициализатор принимает кодируемый символ и количество встречаемых раз.
        """
        self.data = char
        self.right = None
        self.left = None
        self.freq = freq

    def __repr__(self):
        return f"HuffmanNode({self.data}:{self.freq})"

    def __lt__(self, other):
        return self.freq < other.freq

    def __eq__(self, other):
        return self.freq == other.freq

    def __le__(self, other):
        return self.freq <= other.freq

    def __add__(self, other):
        return self.freq + other.freq



"""
Этот модуль содержит линейные структуры данных.
"""


class SinglyLinkedListNode:
    """
    Этот класс представляет собой узел связанного списка (:class:`~SinglyLinkedList`).
    """

    def __init__(self, data):
        """
        Инициализация узла

        :param data: данные, которые хранит узел
        """
        self.data = data
        self.next = None

    def later_node(self, i: int):
        """
        Рекурсивная функция обхода списка. От начала до конца.

        :Сложность: O(i)
        :param i: номер узла последовательности
        :return: i-ый узел последовательности
        """
        if i == 0:
            return self
        if not self.next:
            raise IndexError
        return self.next.later_node(i - 1)

    def __str__(self):
        return str(self.data)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.data == other.data
        return self.data == other

    def __lt__(self, other):
        if hasattr(other, '__lt__'):
            return self.data < other
        return False

    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __gt__(self, other):
        if hasattr(other, '__gt__'):
            return self.data > other
        return False

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)


class SinglyLinkedList:
    """
    Список связан единичными узлами (:class:`~SinglyLinkedListNode`):
    они хранят указатели **только** на следующий узел.

    .. image:: images/singly-linked-list.png
    """

    def __init__(self):
        """
        Инициализатор.
        """
        self.head = None
        self.size = 0

    def __len__(self):
        return self.size

    def __iter__(self):
        trav = self.head
        while trav:
            yield trav.data
            trav = trav.next

    def __str__(self):
        return " -> ".join(map(str, self.__iter__()))

    def __bool__(self):
        return len(self) > 0

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if len(self) != len(other):
                return False

            for node_self, node_other in zip(self, other):
                if node_self != node_other:
                    return False

            return True

        if hasattr(other, '__iter__'):
            return list(self) == list(other)

        return False

    def build(self, data_list: list):
        """
        Преобразует входящий список в LinkedList.

        :Сложность: O(n)
        :param data_list: обычный список
        """
        for data in reversed(data_list):
            self.insert_first(data)

    def get_at(self, i: int):
        """
        Получить данные i-го узла.

        :Сложность: O(i)
        :param i: индекс искомого узла
        :return: данные искомого узла
        """
        node = self.head.later_node(i)
        return node.data

    def __getitem__(self, i):
        return self.get_at(i)

    def set_at(self, i: int, data):
        """
        Меняет данные на data в i-ом узле.

        :Сложность: O(i)
        :param i: индекс узла
        :param data: устанавливаемые данные
        """
        node = self.head.later_node(i)
        node.data = data

    def __setitem__(self, key, value):
        self.set_at(key, value)

    def insert_first(self, data):
        """
        Вставляет узел на 0-ую позицию.

        :Сложность: O(1)
        :param data: Данные, которые будут вставлены в список
        """
        new_node = SinglyLinkedListNode(data)
        new_node.next = self.head
        self.head = new_node
        self.size += 1

    def delete_first(self):
        """
        Удаляет первый узел списка.

        :Сложность: O(1)
        :return: удаляемый узел списка
        """
        temp = self.head.data
        self.head = self.head.next
        self.size -= 1
        return temp

    def insert_at(self, i: int, data):
        """
        Вставляет новый узел на i-ую позицию.

        :Сложность: O(i)
        :param i: индекс позиции нового узла
        :param data: данные нового узла
        """
        if i == 0:
            self.insert_first(data)
            return

        prev_node = self.head.later_node(i - 1)
        new_node = SinglyLinkedListNode(data)
        new_node.next = prev_node.next
        prev_node.next = new_node
        self.size += 1

    def delete_at(self, i: int):
        """
        Удаляет узел на i-ой позиции.

        :Сложность: O(i)
        :param i: индекс узла
        :return: данные удаляемого узла
        """
        if i == 0:
            return self.delete_first()

        prev_node = self.head.later_node(i - 1)
        data = prev_node.next.data
        prev_node.next = prev_node.next.next
        self.size -= 1
        return data

    def insert_last(self, data):
        """
        Добавляет узел в конец списка.

        :Сложность: O(n)
        :param data: данные нового узла
        """
        self.insert_at(len(self), data)

    def delete_last(self):
        """
        Удаляет последний узел.

        :Сложность: O(n)
        :return: данные удаляемого узла
        """
        return self.delete_at(len(self) - 1)

    def search(self, data):
        """
        Поиск узла.

        :Сложность: O(n)
        :param data: искомое значение
        :return: индекс узла, либо -1, если узел не найден
        """
        for item in enumerate(self):
            if item[1] == data:
                return item[0]
        return -1

    def consists(self, data):
        """
        Содержит ли список узел.

        :Сложность: O(n)
        :param data: искомое значение
        :return: True - если содержит, иначе False
        """
        return self.search(data) != -1


class DoublyLinkedListNode(SinglyLinkedListNode):
    """
    Этот класс представляет собой узел двойного связанного списка
    (:class:`~DoublyLinkedList`).
    Он занимает в два раза больше памяти, потому что имеет указатель как и на следующий,
    так и на предыдущий узел.
    """

    def __init__(self, data):
        """
        Инициализация узла

        :param data: данные, которые хранит узел
        """
        self.previous = None
        super().__init__(data)

    def earlier_node(self, i: int):
        """
        Рекурсивная функция обхода списка от конца до начала.

        :Сложность: O(i)
        :param i: номер узла последовательности **от конца**
        :return: i-ый узел последовательности **от конца**
        """
        if i == 0:
            return self
        if not self.previous:
            raise IndexError
        return self.previous.earlier_node(i - 1)


class DoublyLinkedList(SinglyLinkedList):
    """
    Список связан двойными узлами (:class:`~DoublyLinkedListNode`):
    они хранят указатели как на предыдущий узел, так и на следующий.

    .. image:: images/doubly-linked-list.png
    """

    def __init__(self):
        """
        Инициализатор.
        """
        super().__init__()
        self.tail = None

    # __len__ и __iter__ из наследуемого класса

    def build(self, data_list: list):
        """
        Преобразует входящий список в LinkedList.

        :Сложность: O(n)
        :param data_list: обычный список
        """
        for data in data_list:
            self.insert_last(data)

    def get_at(self, i: int):
        """
        Получить данные i-го узла.

        :Сложность: O(i/2)
        :param i: индекс искомого узла
        :return: данные искомого узла
        """
        if len(self) - i - 1 < i:
            node = self.tail.earlier_node(len(self) - i - 1)
        else:
            node = self.head.later_node(i)
        return node.data

    def set_at(self, i: int, data):
        """
        Меняет данные на data в i-ом узле.

        :Сложность: O(i/2)
        :param i: индекс узла
        :param data: устанавливаемые данные
        """
        if len(self) - i - 1 < i:
            node = self.tail.earlier_node(len(self) - i - 1)
        else:
            node = self.head.later_node(i)
        node.data = data

    def insert_first(self, data):
        """
        Вставляет узел на 0-ую позицию.

        :Сложность: O(1)
        :param data: Данные, которые будут вставлены в список
        """
        new_node = DoublyLinkedListNode(data)
        new_node.next = self.head
        self.head = new_node
        self.size += 1

        if self.size == 1:
            self.tail = new_node
            return

        new_node.next.previous = new_node

    def insert_last(self, data):
        """
        Вставляет узел на последнюю позицию.

        :Сложность: O(1)
        :param data: Данные, которые будут вставлены в список
        """
        new_node = DoublyLinkedListNode(data)
        new_node.previous = self.tail
        self.tail = new_node
        self.size += 1

        if self.size == 1:
            self.head = new_node
            return

        new_node.previous.next = new_node

    def delete_first(self):
        """
        Удаляет первый узел списка.

        :Сложность: O(1)
        :return: удаляемый узел списка
        """
        data = self.head.data
        self.head = self.head.next
        if self.head:
            self.head.previous = None
        self.size -= 1
        return data

    def delete_last(self):
        """
        Удаляет последний узел списка.

        :Сложность: O(1)
        :return: удаляемый узел списка
        """

        data = self.tail.data
        self.tail = self.tail.previous
        if self.tail:
            self.tail.next = None
        self.size -= 1
        return data

    def insert_at(self, i: int, data):
        """
        Вставляет новый узел на i-ую позицию.

        :Сложность: ~O(i/2)
        :param i: индекс позиции нового узла
        :param data: данные нового узла
        """
        if i == 0:
            self.insert_first(data)
            return
        if i == len(self):
            self.insert_last(data)
            return

        if i <= len(self) - i - 2:
            prev_node = self.head.later_node(i - 1)
        else:
            prev_node = self.tail.earlier_node(len(self) - i)

        new_node = DoublyLinkedListNode(data)
        new_node.next = prev_node.next
        prev_node.next = new_node
        new_node.previous = prev_node
        new_node.next.previous = new_node
        self.size += 1

    def delete_at(self, i: int):
        """
        Удаляет узел на i-ой позиции.

        :Сложность: ~O(i/2)
        :param i: индекс узла
        :return: данные удаляемого узла
        """
        if i == 0:
            return self.delete_first()
        if i == len(self) - 1:
            self.delete_last()

        if i <= len(self) - i - 2:
            prev_node = self.head.later_node(i - 1)
        else:
            prev_node = self.tail.earlier_node(len(self) - i)

        data = prev_node.next.data
        prev_node.next.next.prev = prev_node
        prev_node.next = prev_node.next.next
        self.size -= 1
        return data


class StaticArray:
    """
    Обычный нерасширяемый список

    .. image:: images/static-array.png
        :width: 300px
    """

    def __init__(self, n: int):
        """
        Инициализация статичного списка.

        :param n: Длина списка
        """
        self.data = [None] * n

    def __iter__(self):
        for item in self.data:
            yield item

    def __bool__(self):
        return len(self.data) > 0

    def __str__(self):
        return self.data.__str__()

    def __eq__(self, other):
        if hasattr(other, '__iter__'):
            return list(self) == list(other)

        return False

    def get_at(self, i: int):
        """
        Получить элемент.

        :Сложность: O(1)
        :param i: Индекс элемента
        :return: Данные под i-ым индексом
        """
        if not 0 <= i < len(self.data):
            raise IndexError
        return self.data[i]

    def __getitem__(self, i):
        return self.get_at(i)

    def set_at(self, i: int, item):
        """
        Установить значение элемента.

        :Сложность: O(1)
        :param i: Индекс элемента
        :param item: Значение элемента
        """
        if not 0 <= i < len(self.data):
            raise IndexError
        self.data[i] = item

    def __setitem__(self, key, value):
        self.set_at(key, value)

    def search(self, item):
        """
        Поиск данных.

        :Сложность: O(n)
        :param item: Искомые данные
        :return: Индекс искомых данных
        """
        i = 0
        for data in self.data:
            if data == item:
                return i
            i += 1
        return -1


class Stack:
    """
    **Last in - First out** список (LIFO). Альтернативное название: стак/стэк.
    Работает так же, как и стопка тарелок.

    .. image:: images/stack.png
        :width: 400px
    """

    def __init__(self):
        """
        Инициализатор. Стак работает с помощью :class:`~SinglyLinkedList`,
        чтобы всё работало со сложностью алгоритма **O(1)**, вместо O(n).
        Такое решение было принято в связи с тем,
        что стаки работают на *сдвигах* и с *крайними элементами списка*.
        Если вы хотите работать со
        стаком как просто со списком, обращайтесь к параметру `data`:
        там находится :class:`~SinglyLinkedList`, на котором всё работает.
        """
        self.data = SinglyLinkedList()

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return self.data.__iter__()

    def __bool__(self):
        return len(self.data) > 0

    def __str__(self):
        return self.data.__str__()

    def __getitem__(self, i):
        return self.data.__getitem__(i)

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.data == other.data
        if hasattr(other, '__iter__'):
            return list(self) == list(other)
        return False

    def push(self, item):
        """
        Положить наверх стопки новый элемент.

        :Сложность: O(1)
        :param item: Значение элемента, которого мы хотим положить наверх стопки
        """
        self.data.insert_first(item)

    def pop(self):
        """
        Удаляет верхний элемент.

        :Сложность: O(1)
        :return: Верхний элемент, который мы удаляем
        """
        return self.data.delete_first()

    def peek(self):
        """
        Узнать верхний элемент.

        :Сложность: O(1)
        :return: Значение верхнего элемента
        """
        return self.data.get_at(0)

    def peek_bottom(self):
        """
        Узнать нижний элемент.

        :Сложность: O(n)
        :return: Значение нижнего элемента
        """
        return self.data.get_at(len(self))


class Queue:
    """
    **First in - First out** список (FIFO). Альтернативное название: очередь.
    Работает так же, как очередь в пивнушке.

    .. image:: images/queue.png
        :width: 400px
    """

    def __init__(self):
        """
        Инициализатор. Очередь работает с помощью :class:`~DoublyLinkedList`,
        чтобы всё работало со сложностью алгоритма **O(1)**, вместо O(n).
        Такое решение было принято в связи с тем, что очереди работают на *сдвигах*
        и с *крайними элементами списка*. Если вы хотите работать с очередью как
        просто со списком, обращайтесь к параметру `data`: там находится
        :class:`~DoublyLinkedList`, на котором всё работает.
        """
        self.data = DoublyLinkedList()

    def __len__(self):
        return self.data.__len__()

    def __iter__(self):
        return self.data.__iter__()

    def __bool__(self):
        return len(self.data) > 0

    def __str__(self):
        return self.data.__str__()

    def __getitem__(self, i):
        return self.data.__getitem__(i)

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.data == other.data
        if hasattr(other, '__iter__'):
            return list(self) == list(other)
        return False

    def enqueue(self, item):
        """
        Поставить в очередь предмет. (Сделать последним элементом *item*)

        :Сложность: O(1)
        :param item: Значение элемента очереди, который мы ставим
        """
        self.data.insert_last(item)

    def push(self, item):
        """
        Поставить в очередь предмет. (Сделать последним элементом *item*)

        :Сложность: O(1)
        :param item: Значение элемента очереди, который мы ставим
        """
        self.enqueue(item)

    def dequeue(self):
        """
        Вынуть из очереди следующий (первый) элемент.

        :Сложность: O(1)
        :return: Значение элемента, которого мы вынимаем
        """
        return self.data.delete_first()

    def pop(self):
        """
        Вынуть из очереди следующий (первый) элемент.

        :Сложность: O(1)
        :return: Значение элемента, которого мы вынимаем
        """
        return self.dequeue()

    def peek_front(self):
        """
        Узнать первый элемент очереди. Он же голова, он же следующий элемент очереди.

        :Сложность: O(1)
        :return: Первый элемент очереди
        """
        return self.data.get_at(0)

    def peek_back(self):
        """
        Узнать последний элемент очереди. Он же хвост, он же конечный элемент очереди.

        :Сложность: O(1)
        :return: Последний элемент очереди
        """
        return self.data.get_at(len(self) - 1)


class Deque(Queue):
    """
    Двусторонняя очередь.

    .. image:: images/deque.jpg
    """
    def push_front(self, item):
        """
        Поставить новый элемент в начало по порядку.

        :Сложность: O(1)
        :param item: значение нового элемента.
        """
        self.data.insert_first(item)

    def push_back(self, item):
        """
        Поставить новый элемент в конец по порядку.

        :Сложность: O(1)
        :param item: значение нового элемента.
        """
        super().enqueue(item)

    def pop_front(self):
        """
        Выбрасывает из очереди первый элемент по порядку.

        :Сложность: O(1)
        :return: значение выбрасываемого элемента
        """
        return super().dequeue()

    def pop_back(self):
        """
        Выбрасывает из очереди последний элемент по порядку.

        :Сложность: O(1)
        :return: значение выбрасываемого элемента
        """
        return self.data.delete_last()

    def enqueue(self, item):
        """
        **Для работы как с обычной очередью:** поставить в очередь предмет.
        (Сделать последним элементом *item*)

        :Сложность: O(1)
        :param item: Значение элемента очереди, который мы ставим
        """
        super().enqueue(item)

    def dequeue(self):
        """
        **Для работы как с обычной очередью:** вынуть из очереди следующий (первый) элемент.

        :Сложность: O(1)
        :return: Значение элемента, которого мы вынимаем
        """
        return super().dequeue()



"""
Этот модуль содержит кучу и приоритетную очередь,
которая, в свою очередь, работает на куче.
"""


from math import floor
from .my_functions import swap


class MinHeap:
    """
    Куча минимума - это бинарное дерево, где ключ каждого узла всегда меньше
    или равен ключам его детей. Эта структура данных позволяет эффективно
    вставлять, удалять и находить минимальный элемент в куче.

    .. image:: images/heap.png
    """

    def __init__(self, arr=None):
        """
        Инициализировать новую кучу минимума.

        :Сложность: O(N log N), где n - количество элементов в arr.
        :param arr: Необязательный итерируемый объект, содержащий начальные элементы кучи.
        """
        if arr is None:
            arr = []
        elif not isinstance(arr, list) and hasattr(arr, '__iter__'):
            arr = list(arr)

        self.heap = arr.copy()

        if len(arr) <= 1:
            return

        for i in range(len(self) // 2, -1, -1):
            self.heapify_down(i)

    def __len__(self):
        """
        Возвращает количество элементов в куче.

        :Сложность: O(1).
        :return: Количество элементов в куче.
        """
        return len(self.heap)

    def __iter__(self):
        """
        Перебирает элементы кучи в порядке их приоритета.

        :Сложность: O(1).
        :return: Итератор по элементам кучи.
        """
        return self.heap.__iter__()

    def get_item(self, index):
        """
        Возвращает элемент по указанному индексу.

        :Сложность: O(1).
        :param index: Индекс элемента для возврата.
        :return: Элемент по указанному индексу.
        """
        return self.heap[index]

    def __getitem__(self, item):
        return self.get_item(item)

    @staticmethod
    def get_left_child_index(index):
        """
        Возвращает индекс левого потомка указанного узла.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: Индекс левого потомка узла.
        """
        return 2 * index + 1

    @staticmethod
    def get_right_child_index(index):
        """
        Возвращает индекс правого потомка указанного узла.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: Индекс правого потомка узла.
        """
        return 2 * index + 2

    @staticmethod
    def get_parent_index(index):
        """
        Возвращает индекс родителя указанного узла.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: Индекс родителя узла.
        """
        return floor((index - 1) / 2)

    def count(self, item):
        """
        Возвращает количество вхождений указанного элемента в кучу.

        :Сложность: O(n), где n - количество элементов в куче.
        :param item: Элемент для подсчета.
        :return: Количество вхождений элемента в кучу.
        """
        return self.heap.count(item)

    def has_left_child(self, index) -> bool:
        """
        Проверяет, есть ли у указанного узла левый потомок.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: True, если узел имеет левого потомка, иначе False.
        """
        return self.get_left_child_index(index) < len(self)

    def has_right_child(self, index) -> bool:
        """
        Проверяет, есть ли у указанного узла правый потомок.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: True, если узел имеет правого потомка, иначе False.
        """
        return self.get_right_child_index(index) < len(self)

    def has_parent(self, index) -> bool:
        """
        Проверяет, есть ли у указанного узла родитель.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: True, если узел имеет родителя, иначе False.
        """
        return self.get_parent_index(index) >= 0

    def left_child(self, index):
        """
        Возвращает левого потомка указанного узла.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: Левый потомок узла.
        """
        return self.heap[self.get_left_child_index(index)]

    def right_child(self, index):
        """
        Возвращает правого потомка указанного узла.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: Правый потомок узла.
        """
        return self.heap[self.get_right_child_index(index)]

    def parent(self, index):
        """
        Возвращает родителя указанного узла.

        :Сложность: O(1).
        :param index: Индекс узла.
        :return: Родитель узла.
        """
        return self.heap[self.get_parent_index(index)]

    def peek(self):
        """
        Возвращает минимальный элемент в куче без его удаления.

        :Сложность: O(1).
        :return: Минимальный элемент в куче.
        """
        if len(self) == 0:
            raise IndexError
        return self.heap[0]

    def poll(self, index=None):
        """
        Удаляет и возвращает минимальный элемент из кучи.

        :Сложность: O(log n), где n - количество элементов в куче.
        :param index: Индекс элемента для удаления. Если не указан, удаляется минимальный элемент.
        :return: Минимальный элемент в куче.
        """
        if len(self) == 0:
            raise IndexError

        if index is None:
            index = 0

        item = self.heap[index]
        self.heap[index] = self.heap[len(self) - 1]
        self.heap.pop()
        self.heapify_down(index)
        return item

    def add(self, item):
        """
        Добавляет элемент в кучу.

        :Сложность: O(log n), где n - количество элементов в куче.
        :param item: Элемент для добавления.
        """
        self.heap.append(item)
        self.heapify_up()

    def heapify_down(self, index=None):
        """
        Перестраивает элементы в куче вниз от указанного узла.

        :Сложность: O(log n), где n - количество элементов в куче.
        :param index: Индекс узла для начала. Если не указан, используется корень кучи.
        """
        if len(self) <= 1:
            return

        if index is None:
            index = 0

        while self.has_left_child(index):
            min_child_index = self.get_left_child_index(index)

            if self.has_right_child(index) and self.right_child(index) < self.left_child(index):
                min_child_index = self.get_right_child_index(index)

            if self[min_child_index] > self[index]:
                break

            swap(self.heap, index, min_child_index)
            index = min_child_index

    def heapify_up(self, index=None):
        """
        Перестраивает элементы в куче вверх от указанного узла.

        :Сложность: O(log n), где n - количество элементов в куче.
        :param index: Индекс узла для начала. Если не указан, используется последний элемент кучи.
        """
        if len(self) <= 1:
            return

        if index is None:
            index = len(self) - 1

        while self.has_parent(index) and self.parent(index) > self.heap[index]:
            p_i = self.get_parent_index(index)
            swap(self.heap, index, p_i)
            index = p_i


class PriorityQueueNode:
    """
    Узел приоритетной очереди.
    """
    def __init__(self, val, priority):
        self.val = val
        self.priority = priority

    def __lt__(self, other):
        if isinstance(other, PriorityQueueNode):
            return self.priority < other.priority
        if isinstance(other, tuple) and len(other) == 2:
            return self.priority < other[1]

        raise TypeError

    def __eq__(self, other):
        if isinstance(other, PriorityQueueNode):
            return self.priority == other.priority
        if isinstance(other, tuple) and len(other) == 2:
            return self.priority == other[1]

        raise TypeError

    def __le__(self, other):
        return self < other or self == other

    def __repr__(self):
        return f"({self.val}, {self.priority})"


class PriorityQueue(MinHeap):
    """
    Приоритетная очередь, реализованная на основе мин-кучи.
    Узлы двигаются в очереди по приоритету.
    """

    def enqueue(self, item, priority=0):
        """
        Добавляет элемент в приоритетную очередь с указанным приоритетом.

        :Сложность: O(log n)
        :param item: Элемент для добавления.
        :param priority: Приоритет элемента. (По умолчанию 0 -- самый высокий приоритет)
        """
        self.add(PriorityQueueNode(item, priority))

    def dequeue(self):
        """
        Извлекает и возвращает элемент с наивысшим приоритетом из приоритетной очереди.

        :Сложность: O(log n)
        :return: Самый приоритетный узел.
        """
        return self.poll().val



"""
Этот модуль содержит графы и алгоритмы, связанные с ними.
"""


from abc import ABC, abstractmethod
from .linear import Queue, Stack
from .heap import PriorityQueue


def _get_path(prev_nodes: list, target: int) -> list:
    """
    Находит путь до цели, используя список из индексов предыдущих узлов.

    :param prev_nodes: список из индексов предыдущих узлов
    :param target: целевой узел, до которого мы вычисляем путь
    :return: список -- последовательность индексов узлов
    """
    path = []
    trav = target
    while trav != -1:
        path.append(trav)
        trav = prev_nodes[trav]
    path.reverse()
    return path


def _default_heuristic(from_node: int, to_node: int) -> int:
    return abs(from_node - to_node)


class _GraphNode:
    """
    Узел графа. Нужен для реализации графа через *связанный список*
    """

    def __init__(self, v2: int, weight: int = 1):
        """
        Инициализатор.

        :param v2: куда
        :param weight: вес
        """
        self.v2 = v2
        self.weight = weight
        self.next = None

    def __str__(self):
        return f"(To: {self.v2}, Weight: {self.weight})"

    def __repr__(self):
        return self.__str__()


class _GraphParent(ABC):

    def __init__(self, size: int, directed: bool = True):
        """
        Инициализатор. Создает пустой граф.

        :param size: количество узлов / размер графа
        :param directed: направленный / ненаправленный граф
        """
        self.size = size
        self.directed = directed

    def __len__(self):
        return self.size

    def add_edge(self, v1: int, v2: int, weight: int = 1, repeat: bool = True):
        """
        Добавить грань.

        :param v1: номер 1го узла
        :param v2: номер 2го узла
        :param repeat: служебная переменная для избегания бесконечной рекурсии, \
        используется для ненаправленных графов
        :param weight: вес грани (его значение)
        """
        if not (0 <= v1 < self.size and 0 <= v2 < self.size):
            raise IndexError
        if not self.directed and repeat:
            self.add_edge(v2, v1, weight, False)

    def remove_edge(self, v1: int, v2: int, repeat: bool = True) -> int:
        """
        Убрать грань.

        :param v1: номер 1го узла
        :param v2: номер 2го узла
        :param repeat: служебная переменная для избегания бесконечной рекурсии, \
        используется для ненаправленных графов
        :return: вес удалённой грани
        """
        if not (0 <= v1 < self.size and 0 <= v2 < self.size):
            raise IndexError
        if not self.directed and repeat:
            self.remove_edge(v2, v1, False)

    @abstractmethod
    def print_adjacency(self):
        """
        Вывести в консоль список / матрицу смежности.
        """
        raise NotImplementedError

    @abstractmethod
    def traversal(self, from_node: int, storage_type: object):
        """
        Проходка по графу.

        :Сложность: O(V + E), где V -- количество вершин и E -- количество рёбер
        :param from_node: номер узла, от которого идёт проходка
        :param storage_type: структура, для управления узлами: \
        очередь -- если в ширину, стэк -- если в глубину
        :return: итерационный объект с номерами узлов
        """
        raise NotImplementedError

    def breadth_first_traversal(self, from_node: int):
        """
        Обход графа в ширину. (*BFT -- Breadth First Traversal*).

        .. image:: images/graph-bfs.gif
            :width: 400px

        :Сложность: O(V + E), где V -- количество вершин и E -- количество рёбер
        :param from_node: номер узла, от которого идёт проходка
        :return: итерационный объект с номерами узлов
        """
        return self.traversal(from_node, Queue)

    def depth_first_traversal(self, from_node: int):
        """
        Обход графа в глубину. (*DFT -- Depth First Traversal*).

        .. image:: images/graph-dfs.gif
            :width: 400px

        :Сложность: O(V + E), где V -- количество вершин и E -- количество рёбер
        :param from_node: номер узла, от которого идёт проходка
        :return: итерационный объект с номерами узлов
        """
        return self.traversal(from_node, Stack)

    @abstractmethod
    def dijkstra(self, from_node: int, to_node: int) -> (int, list):
        """
        Алгоритм Дейкстры. Находит самый короткий путь между двумя узлами.

        **Не поддерживает отрицательные веса.**

        :Сложность: O((V + E) log V), где V - количество вершин
        :param from_node: индекс первого узла (откуда проложить маршрут)
        :param to_node: индекс второго узла (куда проложить маршрут)
        :returns: tuple (int, list)
            - Первый элемент -- это минимальная длина маршрута
            - Второй элемент -- это кратчайший маршрут: последовательность индексов, \
            которые нужно посетить
        """
        raise NotImplementedError


    @abstractmethod
    def bellman_ford(self, from_node: int, to_node: int) -> (int, list):
        """
        Алгоритм Беллмана-Форда. Находит самый короткий путь между двумя узлами.

        **Не поддерживает убывающие циклы. Поддерживает отрицательные веса.**

        :Сложность: O(V * E), где V - количество вершин, а E - количество рёбер
        :param from_node: индекс первого узла (откуда проложить маршрут)
        :param to_node: индекс второго узла (куда проложить маршрут)
        :returns: tuple (int, list)
            - Первый элемент -- это минимальная длина маршрута
            - Второй элемент -- это кратчайший маршрут: последовательность индексов, \
            которые нужно посетить
        """
        raise NotImplementedError


    @abstractmethod
    def a_star(self, from_node: int, to_node: int, heuristic: callable) -> (int, list):
        """
        Алгоритм A*. Находит самый короткий путь между двумя узлами.
        Этот алгоритм является усовершенствованным алгоритмом Дейкстры.
        Он не рассматривает все возможные пути.

        .. image :: images/a_star.gif

        :Сложность: O((V + E) log V), где V - количество вершин, \
        а E - количество рёбер в худшем случае
        :param from_node: индекс первого узла (откуда проложить маршрут)
        :param to_node: индекс второго узла (куда проложить маршрут)
        :param heuristic: функция эвристики для оценки стоимости пути (h(n))
        :returns: tuple (int, list)
            - Первый элемент -- это минимальная длина маршрута
            - Второй элемент -- это кратчайший маршрут: последовательность индексов, \
            которые нужно посетить
        """
        raise NotImplementedError


class ListAdjacency(_GraphParent):
    """
    Реализация графа через список смежности.
    """
    def __init__(self, size: int, directed: bool = True):
        super().__init__(size, directed)
        self.list = [None] * size

    def add_edge(self, v1: int, v2: int, weight: int = 1, repeat: bool = True):
        super().add_edge(v1, v2, weight, repeat)
        node = _GraphNode(v2, weight)
        node.next = self.list[v1]
        self.list[v1] = node

    def remove_edge(self, v1: int, v2: int, repeat: bool = True) -> int:
        super().remove_edge(v1, v2, repeat)
        head = self.list[v1]
        prev = None
        while head and head.v2 != v2:
            prev = head
            head = head.next

        if not head:
            return -1
        tmp = head.weight
        if prev:
            prev.next = head.next
        else:
            self.list[v1] = head.next

        if repeat:
            return tmp

        return 0

    def print_adjacency(self):
        """
        Вывести в консоль список смежности.
        """
        for i in range(self.size):
            v1 = self.list[i]
            print("From:", i, end="")
            while v1:
                print(' -> ', v1, sep='', end='')
                v1 = v1.next
            print()

    def dijkstra(self, from_node: int, to_node: int) -> (int, list):
        visited = [False] * self.size
        distance = [float('inf')] * self.size
        parent = [-1] * self.size
        distance[from_node] = 0

        h = PriorityQueue()
        h.enqueue(from_node, 0)

        while len(h) > 0:
            current_node = h.dequeue()
            if visited[current_node]:
                continue

            visited[current_node] = True
            node = self.list[current_node]
            while node:
                if node.weight < 0:
                    raise ValueError("Алгоритм Дейкстры не поддерживает отрицательные рёбра.")

                # перебор соседей
                if not visited[node.v2] and distance[current_node] + node.weight \
                        < distance[node.v2]:
                    distance[node.v2] = distance[current_node] + node.weight
                    parent[node.v2] = current_node
                    h.enqueue(node.v2, node.weight)
                node = node.next

        # построение маршрута
        if distance[to_node] == float('inf'):
            # нет маршрута
            return -1, []
        return distance[to_node], _get_path(parent, to_node)

    def a_star(self, from_node: int, to_node: int, heuristic: callable = None) -> (int, list):
        if heuristic is None:
            heuristic = _default_heuristic

        visited = [False] * self.size
        distance = [float('inf')] * self.size
        parent = [-1] * self.size
        distance[from_node] = 0

        h = PriorityQueue()
        h.enqueue(from_node, 0)

        while len(h) > 0:
            current_node = h.dequeue()
            if visited[current_node]:
                continue

            # Если мы достигли целевого узла, выходим из цикла
            if current_node == to_node:
                break

            visited[current_node] = True
            node = self.list[current_node]
            while node:
                # перебор соседей
                if not visited[node.v2]:
                    g = distance[current_node] + node.weight
                    if g < distance[node.v2]:
                        distance[node.v2] = g
                        parent[node.v2] = current_node
                        f_score = g + heuristic()
                        h.enqueue(node.v2, f_score)
                node = node.next

        # Построение маршрута
        if distance[to_node] == float('inf'):
            # Нет маршрута
            return -1, []

        return distance[to_node], _get_path(parent, to_node)

    def bellman_ford(self, from_node: int, to_node: int) -> (int, list):
        distance = [float('inf')] * self.size
        parent = [-1] * self.size
        distance[from_node] = 0

        for _ in range(self.size - 1):
            for i in range(self.size):
                node = self.list[i]
                while node:
                    if distance[i] != float('inf') and \
                            distance[i] + node.weight < distance[node.v2]:
                        distance[node.v2] = distance[i] + node.weight
                        parent[node.v2] = i
                    node = node.next

        # Проверяем наличие отрицательных циклов
        for i in range(self.size):
            node = self.list[i]
            while node:
                if distance[i] != float('inf') and distance[i] + node.weight < distance[node.v2]:
                    raise ValueError("Граф содержит отрицательный цикл")
                node = node.next

        if distance[to_node] == float('inf'):
            return -1, []

        return distance[to_node], _get_path(parent, to_node)

    def traversal(self, from_node: int, storage_type: type):
        storage = storage_type()
        storage.push(from_node)
        visited = set()
        visited.add(from_node)
        while storage:
            node_num = storage.pop()

            yield node_num

            trav = self.list[node_num]
            while trav:
                if trav.v2 not in visited:
                    storage.push(trav.v2)
                    visited.add(trav.v2)
                trav = trav.next


class MatrixAdjacency(_GraphParent):
    """
    Реализация графа через матрицу смежности.
    """
    def __init__(self, size: int, directed: bool = True):
        super().__init__(size, directed)
        self.matrix = []
        for _ in range(size):
            self.matrix.append([0 for _ in range(size)])

    def add_edge(self, v1: int, v2: int, weight: int = 1, repeat: bool = True) -> None:
        super().add_edge(v1, v2, weight, repeat)
        self.matrix[v1][v2] = weight

    def remove_edge(self, v1: int, v2: int, repeat: bool = True) -> int:
        super().remove_edge(v1, v2, repeat)
        tmp = self.matrix[v1][v2]
        self.matrix[v1][v2] = 0
        if repeat:
            return tmp
        return 0

    def print_adjacency(self):
        """
        Вывести матрицу в консоль.
        """
        for row in self.matrix:
            for column in row:
                print(column, end=" ")
            print()

    def dijkstra(self, from_node: int, to_node: int) -> (int, list):
        visited = [False] * self.size
        distance = [float('inf')] * self.size
        parent = [-1] * self.size
        distance[from_node] = 0

        h = PriorityQueue()
        h.enqueue(from_node, 0)

        while len(h) > 0:
            current_node = h.dequeue()
            if visited[current_node]:
                continue

            visited[current_node] = True
            for neighbour in range(self.size):
                if self.matrix[current_node][neighbour] == 0:
                    continue
                new_distance = self.matrix[current_node][neighbour] + distance[current_node]
                if new_distance < distance[neighbour]:
                    distance[neighbour] = new_distance
                    parent[neighbour] = current_node
                    h.enqueue(neighbour, new_distance)

        return distance[to_node], _get_path(parent, to_node)

    def bellman_ford(self, from_node: int, to_node: int) -> (int, list):
        distance = [float('inf')] * self.size
        parent = [-1] * self.size
        distance[from_node] = 0

        for _ in range(self.size - 1):
            for v1 in range(self.size):
                for v2 in range(self.size):
                    weight = self.matrix[v1][v2]
                    if weight != 0 and distance[v1] != float('inf') \
                            and distance[v1] + weight < distance[v2]:
                        distance[v2] = distance[v1] + weight
                        parent[v2] = v1

        # Проверяем наличие отрицательных циклов
        for v1 in range(self.size):
            for v2 in range(self.size):
                weight = self.matrix[v1][v2]
                if weight != 0 and distance[v1] != float('inf') \
                        and distance[v1] + weight < distance[v2]:
                    raise ValueError("Граф содержит отрицательный цикл")

        if distance[to_node] == float('inf'):
            return -1, []

        return distance[to_node], _get_path(parent, to_node)

    def a_star(self, from_node: int, to_node: int, heuristic: callable = None) -> (int, list):
        if heuristic is None:
            heuristic = _default_heuristic

        visited = [False] * self.size
        distance = [float('inf')] * self.size
        parent = [-1] * self.size
        distance[from_node] = 0

        h = PriorityQueue()
        h.enqueue(from_node, heuristic(from_node, to_node))

        while len(h) > 0:
            current_node = h.dequeue()
            if current_node == to_node:
                break
            if visited[current_node]:
                continue

            visited[current_node] = True

            for neighbour in range(self.size):
                if self.matrix[current_node][neighbour] == 0:
                    continue
                g = self.matrix[current_node][neighbour] + distance[current_node]
                if g < distance[neighbour]:
                    distance[neighbour] = g
                    parent[neighbour] = current_node
                    f_score = g + heuristic(neighbour, to_node)
                    h.enqueue(neighbour, f_score)

        if distance[to_node] == float('inf'):
            return -1, []

        return distance[to_node], _get_path(parent, to_node)

    def traversal(self, from_node: int, storage_type: type):
        storage = storage_type()
        storage.push(from_node)
        visited = set()
        visited.add(from_node)
        while storage:
            node_num = storage.pop()

            yield node_num

            for i in range(self.size):
                if self.matrix[node_num][i] > 0 and i not in visited:
                    storage.push(i)
                    visited.add(i)
    


"""
Дополнительный модуль, который содержит некоторые функции.
"""


def swap(arr, i, j):
    """
    Меняет местами два элемента в массиве.
    """
    arr[i], arr[j] = arr[j], arr[i]


def is_sorted(arr):
    """
    Проверяет, является ли массив отсортированным по возрастанию.
    """
    if len(arr) <= 1:
        return True

    # Проверяем список на возрастание
    for i in range(1, len(arr)):
        if arr[i] < arr[i - 1]:
            return False

    return True

def merge(arr1, arr2) -> list:
    """
    Сливает два массива в один по возрастанию значений.
    """
    arr = []
    i = j = 0

    while i < len(arr1) and j < len(arr2):
        if arr1[i] < arr2[j]:
            arr.append(arr1[i])
            i += 1
        else:
            arr.append(arr2[j])
            j += 1

    arr.extend(arr1[i:])
    arr.extend(arr2[j:])

    return arr



import setuptools

VERSION = "1.1"

NAME = "slav-eeik-datastructures"

INSTALL_REQUIRES = []

EXTRAS_REQUIRE = {
    "dev": [
        "sphinx",
        "sphinx-rtd-theme",
        "pytest",
    ]
}

setuptools.setup(
    name=NAME,
    version=VERSION,
    description="Simple datastructures library with documentation.",
    url="https://github.com/zhukovrost/datastructures",
    project_urls={
        "Source Code": "https://github.com/zhukovrost/datastructures",
        "Documentation": "https://zhukovrost.github.io/datastructures/",
    },
    author="Zhukov Rostislav",
    author_email="2rosf.zhukov@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
    ],

    python_requires=">=3.10",

    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    packages=setuptools.find_packages(),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
)



"""
Этот модуль содержит реализации различных бинарных и не только деревьев.
"""


from abc import ABC
from collections import defaultdict
from .nodes import BinaryNode, SearchNode, SegmentNode, \
    AVLNode, TrieNode, RedBlackNode, TwoThreeTreeNode, HuffmanNode
from ..heap import MinHeap


class _BinaryTree(ABC):
    def __init__(self, tree_node_type=BinaryNode):
        """
        Инициализатор.

        :param tree_node_type: Тип узлов дерева. По умолчанию :class:`~BinaryTreeNode`
        """
        self.root = None
        self.size = 0
        self.tree_node_type = tree_node_type

    def __len__(self):
        return self.size

    def __iter__(self):
        if self.root:
            for node in self.root.subtree_iter():
                yield node

    def __repr__(self):
        if self.root:
            return f"BinaryTree({[node.__repr__() for node in self.root.subtree_iter()]})"
        return "BinaryTree([])"


class SearchTree(_BinaryTree):
    """
    Это то же самое бинарное дерево, но значения узлов будут возрастать в порядке обхода.
    Повторяющиеся элементы пропадают. По другому это дерево можно назвать бинарное
    дерево-множество или дерево бинарного поиска.

    .. image:: /images/bst.png
        :width: 500px

    Если обойти это дерево по порядку (:class:`~BinaryTreeNode.subtree_iter`),
    то на выходе мы получим узлы в порядке возрастания:
    **2 4 6 8 9 10 11 12 14 16 18**

    .. image:: images/BST_complexity.png
    """

    def __init__(self, tree_node_type=SearchNode):
        """
        Инициализатор дерева с типом узлов :class:`~BSTNode`.
        """
        super().__init__(tree_node_type)

    def build(self, arr: list):
        """
        Строит дерево бинарного поиска из входящего списка.

        :Сложность: O(nlog n)
        :param arr: входящий список значений узлов
        """
        for item in arr:
            self.insert(item)

    def find_min(self):
        """
        Минимальный элемент - первый, поэтому вызываем :class:`~BSTNode.subtree_first`

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: значение минимального узла
        """
        if self.root:
            return self.root.subtree_first().data
        return None

    def find_max(self):
        """
        Максимальный элемент - последний, поэтому вызываем :class:`~BSTNode.subtree_last`

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :return: значение максимального узла
        """
        if self.root:
            return self.root.subtree_last().data
        return None

    def find(self, item: int):
        """
        Найти элемент.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: искомое значение
        :return: значение искомого элемента, если элемент не найден возвращает None
        """
        if self.root:
            node = self.root.subtree_find(item)
            if node is not None:
                return node.data
        return None

    def find_next(self, item: int):
        """
        Найти элемент, идущий после данного.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение элемента, после которого идёт искомый
        :return: значение искомого элемента, если элемент не найден возвращает None
        """
        if self.root:
            node = self.root.subtree_find_next(item)
            if node is not None:
                return node.data
        return None

    def find_prev(self, item: int):
        """
        Найти элемент, идущий перед данным.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение элемента, перед которым идёт искомый
        :return: значение искомого элемента, если элемент не найден возвращает None
        """
        if self.root:
            node = self.root.subtree_find_prev(item)
            if node is not None:
                return node.data
        return None

    def _insert_fix(self, n):
        pass

    def insert(self, item: int) -> bool:
        """
        Вставить узел.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение нового узла
        :return: *True*, если узел был добавлен, иначе *False*
        """
        new_node = self.tree_node_type(item)
        if self.root:
            self.root.subtree_insert(new_node)
            if new_node.parent is None:
                # if the value repeats
                return False
        else:
            self.root = new_node

        self.size += 1
        self._insert_fix(new_node)
        return True

    def _delete_fix(self, x):
        pass

    def delete(self, item: int):
        """
        Удалить узел.

        :Сложность: O(log n), *(если дерево не сбалансировано - O(h))*
        :param item: значение удаляемого узла
        :return: значение удаляемого узла
        """
        assert self.root
        node = self.root.subtree_find(item)
        assert node
        temp = node.subtree_delete()
        self.size -= 1

        if self.size == 0:
            self.root = None

        self._delete_fix(node)

        return temp.data


class RedBlackTree(SearchTree):
    """
    Красно-черное дерево - это самобалансирующееся дерево бинарного поиска.
    Оно поддерживает балансировку за счет следующих правил:

    1. Каждый узел либо красный, либо черный.
    2. Корень всегда черный.
    3. Все листья (NULL узлы) черные.
    4. Оба ребенка каждого красного узла черные.
    5. Любой путь от узла до листьев имеет одинаковое количество черных узлов.

    .. image:: images/redblack.png

    """

    def __init__(self):
        """
        Инициализатор.
        """
        super().__init__(tree_node_type=RedBlackNode)

    def _insert_fix(self, node):
        while node.parent and node.parent.color is RedBlackNode.RED:
            if node.parent == node.grandparent.left:
                uncle = node.uncle
                if uncle and uncle.color is RedBlackNode.RED:
                    node.parent.color = RedBlackNode.BLACK
                    uncle.color = RedBlackNode.BLACK
                    node.grandparent.color = RedBlackNode.RED
                    node = node.grandparent
                else:
                    if node == node.parent.right:
                        node = node.parent
                        self._rotate_left(node)
                    node.parent.color = RedBlackNode.BLACK
                    node.grandparent.color = RedBlackNode.RED
                    self._rotate_right(node.grandparent)
            else:
                uncle = node.uncle
                if uncle and uncle.color is RedBlackNode.RED:
                    node.parent.color = RedBlackNode.BLACK
                    uncle.color = RedBlackNode.BLACK
                    node.grandparent.color = RedBlackNode.RED
                    node = node.grandparent
                else:
                    if node == node.parent.left:
                        node = node.parent
                        self._rotate_right(node)
                    node.parent.color = RedBlackNode.BLACK
                    node.grandparent.color = RedBlackNode.RED
                    self._rotate_left(node.grandparent)
        self.root.color = RedBlackNode.BLACK

    def _delete_fix(self, node):
        while node != self.root and node.color is RedBlackNode.BLACK:
            if node == node.parent.left:
                sibling = node.sibling
                if sibling.color is RedBlackNode.RED:
                    sibling.color = RedBlackNode.BLACK
                    node.parent.color = RedBlackNode.RED
                    self._rotate_left(node.parent)
                    sibling = node.sibling
                if (sibling.left is None or sibling.left.color is RedBlackNode.BLACK) \
                        and (sibling.right is None or sibling.right.color is RedBlackNode.BLACK):
                    sibling.color = RedBlackNode.RED
                    node = node.parent
                else:
                    if sibling.right is None or sibling.right.color is RedBlackNode.BLACK:
                        sibling.left.color = RedBlackNode.BLACK
                        sibling.color = RedBlackNode.RED
                        self._rotate_right(sibling)
                        sibling = node.sibling
                    sibling.color = node.parent.color
                    node.parent.color = RedBlackNode.BLACK
                    if sibling.right:
                        sibling.right.color = RedBlackNode.BLACK
                    self._rotate_left(node.parent)
                    node = self.root
            else:
                sibling = node.sibling
                if sibling.color is RedBlackNode.RED:
                    sibling.color = RedBlackNode.BLACK
                    node.parent.color = RedBlackNode.RED
                    self._rotate_right(node.parent)
                    sibling = node.sibling
                if (sibling.left is None or sibling.left.color is RedBlackNode.BLACK) and \
                        (sibling.right is None or sibling.right.color is RedBlackNode.BLACK):
                    sibling.color = RedBlackNode.RED
                    node = node.parent
                else:
                    if sibling.left is None or sibling.left.color is RedBlackNode.BLACK:
                        sibling.right.color = RedBlackNode.BLACK
                        sibling.color = RedBlackNode.RED
                        self._rotate_left(sibling)
                        sibling = node.sibling
                    sibling.color = node.parent.color
                    node.parent.color = RedBlackNode.BLACK
                    if sibling.left:
                        sibling.left.color = RedBlackNode.BLACK
                    self._rotate_right(node.parent)
                    node = self.root
        node.color = RedBlackNode.BLACK

    def _rotate_left(self, node):
        right_child = node.right
        node.right = right_child.left

        if right_child.left is not None:
            right_child.left.parent = node

        right_child.parent = node.parent

        if node.parent is None:
            self.root = right_child
        elif node == node.parent.left:
            node.parent.left = right_child
        else:
            node.parent.right = right_child

        right_child.left = node
        node.parent = right_child

    def _rotate_right(self, node):
        left_child = node.left
        node.left = left_child.right

        if left_child.right is not None:
            left_child.right.parent = node

        left_child.parent = node.parent

        if node.parent is None:
            self.root = left_child
        elif node == node.parent.right:
            node.parent.right = left_child
        else:
            node.parent.left = left_child

        left_child.right = node
        node.parent = left_child


class SegmentTree(_BinaryTree):
    """
    Бинарное дерево, к элементам которого можно обращаться по индексу (*Sequence Binary Tree*).
    **Индексация идет по порядку обхода дерева**.
    """
    def __init__(self, tree_node_type=SegmentNode):
        super().__init__(tree_node_type)

    def build(self, iterable: list):
        """
        Построить дерево из входящего списка. Но алгоритм не просто поочерёдно вставлять элементы,
        а корнем каждого поддерева является центр среза входящего списка, что способствует балансу
        дерева.

        :Сложность: O(n)
        :param iterable: входящий список
        """
        def build_subtree(_iterable: list, _from: int, _to: int):
            center = (_from + _to) // 2
            root = self.tree_node_type(_iterable[center])
            if _from < center:
                root.left = build_subtree(_iterable, _from, center - 1)
                root.left.parent = root
            if _to > center:
                root.right = build_subtree(_iterable, center + 1, _to)
                root.right.parent = root
            root.subtree_update()
            return root

        self.root = build_subtree(iterable, 0, len(iterable) - 1)
        self.size = self.root.size

    def get_at(self, i):
        """
        Получить i-ый узел.

        :Сложность: O(log n)
        :param i: индекс узла
        :return: значение искомого узла
        """
        assert self.root
        return self.root.subtree_at(i).data

    def __getitem__(self, i):
        return self.get_at(i)

    def set_at(self, i, data):
        """
        Установить значение i-го узла.

        :Сложность: O(log n)
        :param i: индекс узла
        :param data: новое значение узла
        """
        assert self.root
        self.root.subtree_at(i).data = data

    def __setitem__(self, key, value):
        self.set_at(key, value)

    def insert_at(self, i, data):
        """
        Вставить новый узел на i-ую позицию.

        :Сложность: O(log n)
        :param i: индекс нового узла
        :param data: значение нового узла
        """
        new_node = self.tree_node_type(data)
        if i == 0:
            if self.root:
                node = self.root.subtree_first()
                node.subtree_insert_before(new_node)
            else:
                self.root = new_node
        else:
            node = self.root.subtree_at(i - 1)
            node.subtree_insert_after(new_node)
        self.size += 1

    def delete_at(self, i):
        """
        Удаляет i-ый элемент.

        :Сложность: O(log n)
        :param i: индекс удаляемого узла
        :return: значение удаляемого узла
        """
        assert self.root
        node = self.root.subtree_at(i)
        temp = node.subtree_delete()
        if temp is self.root and temp.left is None and temp.right is None:
            self.root = None
        self.size -= 1
        return temp.data

    def delete_first(self):
        """
        Удаляет первый узел.

        :Сложность: O(log n)
        :return: значение удаляемого узла
        """
        return self.delete_at(0)

    def delete_last(self):
        """
        Удаляет последний узел.

        :Сложность: O(log n)
        :return: значение удаляемого узла
        """
        return self.delete_at(len(self) - 1)

    def insert_first(self, data):
        """
        Вставляет новый узел в начало дерева (0 позиция).

        :Сложность: O(log n)
        :param data: значение нового узла
        """
        self.insert_at(0, data)

    def insert_last(self, data):
        """
        Вставляет новый узел в конец дерева (len(self) позиция).

        :Сложность: O(log n)
        :param data: значение нового узла
        """
        self.insert_at(len(self), data)


class AVLTree(SearchTree):
    """
    AVL дерево - бинарное балансирующее дерево поиска.
    Балансировка осуществляется засчёт поворотов дерева.
    """
    def __init__(self, tree_node_type=AVLNode):
        super().__init__(tree_node_type)


class Trie:
    """
    Trie (также известно как префиксное дерево) - это древовидная структура данных,
    которая хранит динамический набор или ассоциативный массив.
    Ключ обычно является строкой, а связанное значение часто является None.
    """

    def __init__(self):
        """
        Инициализирует пустое дерево Trie.
        """
        self.root = TrieNode()

    def insert(self, *words):
        """
        Вставляет слова (строки) в дерево Trie.

        :param words: Переменная длина аргумента списка слов для вставки.
        :raises TypeError: Если слово не является строкой.
        """
        for word in words:
            if not isinstance(word, str):
                raise TypeError("Метод Trie.insert() принимает только строки")
            node = self.root
            for char in word.lower():
                if char not in node:
                    node[char] = TrieNode()
                node = node[char]
            node.is_word = True

    def search(self, word):
        """
        Ищет слово в дереве Trie.

        :param word: Слово для поиска.
        :return: True, если слово найдено, иначе False.
        :raises TypeError: Если слово не является строкой.
        """
        if not isinstance(word, str):
            raise TypeError("Метод Trie.search() принимает только строки")
        node = self.root
        for char in word:
            if char not in node:
                return False
            node = node[char]
        return node.is_word

    def __contains__(self, word):
        """
        Проверяет, содержится ли слово в дереве Trie.

        :param word: Слово для проверки.
        :return: True, если слово есть в дереве Trie, иначе False.
        """
        return self.search(word)

    def __iter__(self):
        """
        Перебирает все слова в дереве Trie.

        :return: Итератор по всем словам в дереве Trie.
        """
        return self.root.get_words()

    def get_words_with_prefix(self, prefix=""):
        """
        Получает все слова в дереве Trie с указанным префиксом.

        :param prefix: Префикс для проверки.
        :return: Список слов с указанным префиксом.
        :raises TypeError: Если префикс не является строкой.
        """
        if not isinstance(prefix, str):
            raise TypeError("Метод Trie.get_words() принимает только строки")
        node = self.root
        for char in prefix:
            if char not in node:
                return []
            node = node[char]
        return list(node.get_words(prefix))


class TwoThreeTree:
    """
    Дерево 2-3 - это B-дерево порядка 3.

    Свойства дерева 2-3:

     - Узлы с двумя дочерними элементами называются 2-узлами. 2-узлы имеют одно \
      значение данных и два дочерних узла.
     - Узлы с тремя детьми называются 3-узлами. \
     3-узлы имеют два значения данных и три дочерних узла.
     - Данные хранятся в отсортированном порядке.
     - Это сбалансированное дерево.
     - Все листовые узлы находятся на одном уровне.
     - Каждый узел может быть либо листом, либо 2-узловым, либо 3-узловым.
     - Вставка всегда выполняется в лист.

    .. image:: images/2-3.jpg

    """
    def __init__(self):
        """
        Инициализатор.
        """
        self.root = None

    def search(self, key, node=None) -> bool:
        """
        Поиск ключа в дереве.

        Выполняет рекурсивный поиск ключа в узлах дерева.
        Если ключ найден, возвращает True, иначе - False.

        :Сложность: O(logN)
        :param key: Ключ, который нужно найти.
        :param node: Узел, с которого начинается поиск (по умолчанию корневой узел).
        :return: True, если ключ найден, иначе False.
        """
        if node is None:
            node = self.root

        if node is None:
            return False

        i = 0
        while i < len(node.keys) and key > node.keys[i]:
            i += 1
        if i < len(node.keys) and key == node.keys[i]:
            return True
        if node.is_leaf():
            return False
        return self.search(key, node.children[i])

    def insert(self, key):
        """
        Вставка ключа в дерево 2-3.

        Если дерево пустое, создается новый корневой узел.
        Если в результате вставки происходит разделение узла,
        то корень обновляется.

        :Сложность: O(log n)
        :param key: Ключ, который нужно вставить.
        """
        if self.root is None:
            self.root = TwoThreeTreeNode([key])
        else:
            new_node, new_key = self._insert(self.root, key)
            if new_node:
                self.root = TwoThreeTreeNode([new_key], [self.root, new_node])

    def _insert(self, node, key):
        if node.is_leaf():
            return self._insert_into_leaf(node, key)
        i = 0
        while i < len(node.keys) and key > node.keys[i]:
            i += 1
        new_node, new_key = self._insert(node.children[i], key)
        if new_node:
            return self._insert_into_internal(node, new_key, new_node)
        return None, None

    def _insert_into_leaf(self, node, key):
        node.keys.append(key)
        node.keys.sort()
        if node.is_full():
            return self._split(node)
        return None, None

    def _insert_into_internal(self, node, key, child):
        node.keys.append(key)
        node.keys.sort()
        i = node.keys.index(key)
        node.children.insert(i + 1, child)
        if node.is_full():
            return self._split(node)
        return None, None

    def _split(self, node):
        mid_key = node.keys[1]
        if node.is_leaf():
            new_node = TwoThreeTreeNode([node.keys[2]])
        else:
            new_node = TwoThreeTreeNode([node.keys[2]], node.children[2:])
        node.keys = [node.keys[0]]
        node.children = node.children[:2] if node.children else []
        return new_node, mid_key


class HuffmanTree(_BinaryTree):
    """
    Дерево Хаффмана - это структура, на которой держится алгоритм сжатия Хаффмана.
    Алгоритм заключается в том, что мы кодируем элементы строки по-своему:
    элементы, которые встречаются чаще, кодируются короче.

    .. image:: images/huffman_tree.png

    """
    def __init__(self, tree_node_type=HuffmanNode):
        super().__init__(tree_node_type)

    @staticmethod
    def get_frequency_map(s: str) -> dict:
        """
        Статический метод для образования словаря частот.

        :param s: строка, для которой нужен словарь частот
        :return: Словарь, где ключи — символы, а значения — их частоты
        """
        res = defaultdict(int)
        for char in s:
            res[char] += 1
        return dict(res)


    def build(self, frequency_map=None):
        """
        Построение дерева Хаффмана из словаря частот символов.

        :param frequency_map: Словарь, где ключи — символы, а значения — их частоты.
        """
        if frequency_map is None:
            frequency_map = {}

        heap = MinHeap([HuffmanNode(char, freq) for char, freq in frequency_map.items()])

        while len(heap) > 1:
            left = heap.poll()
            right = heap.poll()

            merged_node = HuffmanNode(None, left.freq + right.freq)
            merged_node.left = left
            merged_node.right = right

            heap.add(merged_node)

        self.root = heap.poll()  # Устанавливаем корень дерева
        self.size = len(frequency_map)

    def generate_codes(self):
        """
        Генерация кодов Хаффмана для каждого символа на основе дерева.

        :return: Словарь, где ключ — символ, значение — его код Хаффмана.
        """
        codes = {}

        def _generate_codes(node, current_code=""):
            if node is None:
                return

            # Если это листовой узел, сохраняем код
            if node.data is not None:
                codes[node.data] = current_code

            _generate_codes(node.left, current_code + "0")
            _generate_codes(node.right, current_code + "1")

        _generate_codes(self.root)
        return codes

    def encode(self, text):
        """
        Кодирование текста с использованием дерева Хаффмана.

        :param text: Текст, который нужно закодировать.
        :return: Закодированная строка.
        """
        codes = self.generate_codes()
        return ''.join(codes[char] for char in text)

    def decode(self, encoded_text):
        """
        Декодирование строки, используя дерево Хаффмана.

        :param encoded_text: Закодированная строка.
        :return: Декодированная строка.
        """
        decoded_text = []
        current_node = self.root

        for bit in encoded_text:
            if bit == '0':
                current_node = current_node.left
            else:
                current_node = current_node.right

            if current_node.data is not None:  # Листовой узел
                decoded_text.append(current_node.data)
                current_node = self.root

        return ''.join(decoded_text)
